/* gen-script-tab.c - generate pango-script-table.h for pango
 *
 * Author:
 *   Behdad Esfahbod, 2001, 2002, 2004
 *
 * Copyright (C) 2004 Sharif FarsiWeb, Inc
 * Copyright (C) 2001,2002,2004 Behdad Esfahbod
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library, in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA
 * 
 * For licensing issues, contact <license@farsiweb.info>.
 */

#include <glib.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "packtab.h"

#define appname "gen-script-tab"
#define outputname "pango-script-table.h"

#define DIRECTLEN 0x0000

static void
die (
  const char *msg
)
{
  fprintf (stderr, appname ": %s\n", msg);
  exit (1);
}

static void
die2 (
  const char *fmt,
  const char *p
)
{
  fprintf (stderr, appname ": ");
  fprintf (stderr, fmt, p);
  fprintf (stderr, "\n");
  exit (1);
}

static void
die3 (
  const char *fmt,
  unsigned long l,
  const char *p
)
{
  fprintf (stderr, appname ": ");
  fprintf (stderr, fmt, l, p);
  fprintf (stderr, "\n");
  exit (1);
}

struct
{
  const char *unicode_name;
  const char *pango_name;
}
type_names[] =
{
  {"Common",			"PANGO_SCRIPT_COMMON"},
  {"Inherited",			"PANGO_SCRIPT_INHERITED"},
  {"Arabic",			"PANGO_SCRIPT_ARABIC"},
  {"Armenian",			"PANGO_SCRIPT_ARMENIAN"},
  {"Bengali",			"PANGO_SCRIPT_BENGALI"},
  {"Bopomofo",			"PANGO_SCRIPT_BOPOMOFO"},
  {"Cherokee",			"PANGO_SCRIPT_CHEROKEE"},
  {"Coptic",			"PANGO_SCRIPT_COPTIC"},
  {"Cyrillic",			"PANGO_SCRIPT_CYRILLIC"},
  {"Deseret",			"PANGO_SCRIPT_DESERET"},
  {"Devanagari",		"PANGO_SCRIPT_DEVANAGARI"},
  {"Ethiopic",			"PANGO_SCRIPT_ETHIOPIC"},
  {"Georgian",			"PANGO_SCRIPT_GEORGIAN"},
  {"Gothic",			"PANGO_SCRIPT_GOTHIC"},
  {"Greek",			"PANGO_SCRIPT_GREEK"},
  {"Gujarati",			"PANGO_SCRIPT_GUJARATI"},
  {"Gurmukhi",			"PANGO_SCRIPT_GURMUKHI"},
  {"Han",			"PANGO_SCRIPT_HAN"},
  {"Hangul",			"PANGO_SCRIPT_HANGUL"},
  {"Hebrew",			"PANGO_SCRIPT_HEBREW"},
  {"Hiragana",			"PANGO_SCRIPT_HIRAGANA"},
  {"Kannada",			"PANGO_SCRIPT_KANNADA"},
  {"Katakana",			"PANGO_SCRIPT_KATAKANA"},
  {"Khmer",			"PANGO_SCRIPT_KHMER"},
  {"Lao",			"PANGO_SCRIPT_LAO"},
  {"Latin",			"PANGO_SCRIPT_LATIN"},
  {"Malayalam",			"PANGO_SCRIPT_MALAYALAM"},
  {"Mongolian",			"PANGO_SCRIPT_MONGOLIAN"},
  {"Myanmar",			"PANGO_SCRIPT_MYANMAR"},
  {"Ogham",			"PANGO_SCRIPT_OGHAM"},
  {"Old_Italic",		"PANGO_SCRIPT_OLD_ITALIC"},
  {"Oriya",			"PANGO_SCRIPT_ORIYA"},
  {"Runic",			"PANGO_SCRIPT_RUNIC"},
  {"Sinhala",			"PANGO_SCRIPT_SINHALA"},
  {"Syriac",			"PANGO_SCRIPT_SYRIAC"},
  {"Tamil",			"PANGO_SCRIPT_TAMIL"},
  {"Telugu",			"PANGO_SCRIPT_TELUGU"},
  {"Thaana",			"PANGO_SCRIPT_THAANA"},
  {"Thai",			"PANGO_SCRIPT_THAI"},
  {"Tibetan",			"PANGO_SCRIPT_TIBETAN"},
  {"Canadian_Aboriginal",	"PANGO_SCRIPT_CANADIAN_ABORIGINAL"},
  {"Yi",			"PANGO_SCRIPT_YI"},
  {"Tagalog",			"PANGO_SCRIPT_TAGALOG"},
  {"Hanunoo",			"PANGO_SCRIPT_HANUNOO"},
  {"Buhid",			"PANGO_SCRIPT_BUHID"},
  {"Tagbanwa",			"PANGO_SCRIPT_TAGBANWA"},
  {"Braille",			"PANGO_SCRIPT_BRAILLE"},
  {"Cypriot",			"PANGO_SCRIPT_CYPRIOT"},
  {"Limbu",			"PANGO_SCRIPT_LIMBU"},
  {"Osmanya",			"PANGO_SCRIPT_OSMANYA"},
  {"Shavian",			"PANGO_SCRIPT_SHAVIAN"},
  {"Linear_B",			"PANGO_SCRIPT_LINEAR_B"},
  {"Tai_Le",			"PANGO_SCRIPT_TAI_LE"},
  {"Ugaritic",			"PANGO_SCRIPT_UGARITIC"},
  {"New_Tai_Lue",		"PANGO_SCRIPT_NEW_TAI_LUE"},
  {"Buginese",			"PANGO_SCRIPT_BUGINESE"},
  {"Glagolitic",		"PANGO_SCRIPT_GLAGOLITIC"},
  {"Tifinagh",			"PANGO_SCRIPT_TIFINAGH"},
  {"Syloti_Nagri",		"PANGO_SCRIPT_SYLOTI_NAGRI"},
  {"Old_Persian",		"PANGO_SCRIPT_OLD_PERSIAN"},
  {"Kharoshthi",		"PANGO_SCRIPT_KHAROSHTHI"},
};

#define type_names_count (sizeof (type_names) / sizeof (type_names[0]))

static const char *names[type_names_count];

static char
get_type (
  const char *s
)
{
  unsigned int i;

  for (i = 0; i < type_names_count; i++)
    if (!strcmp (s, type_names[i].unicode_name))
      return i;
  die2 ("bidi type name `%s' not found", s);
  return 0;
}

#define table_name "Scr"
#define macro_name "PANGO_GET_SCRIPT"

#define UNICODE_CHARS 0x110000


static signed int table[UNICODE_CHARS];
static char buf[4000];
static char tp[sizeof (buf)];

static void
init (
  void
)
{
  register int i;

  for (i = 0; i < type_names_count; i++)
    names[i] = 0;
  for (i = type_names_count - 1; i >= 0; i--)
    names[i] = type_names[i].pango_name;
}

static void
clear_tab (
  void
)
{
  register gunichar c;

  for (c = 0; c < UNICODE_CHARS; c++)
    table[c] = 0;
}

static void
init_tab_derived_bidi_class_txt (
  void
)
{
  clear_tab ();
}

static void
read_scripts_txt (
  FILE *f
)
{
  unsigned long c, c2, l;

  init_tab_derived_bidi_class_txt ();

  l = 0;
  while (fgets (buf, sizeof buf, f))
    {
      int i;
      register char typ;
      const char *s = buf;

      l++;

      while (*s == ' ')
	s++;

      if (s[0] == '#' || s[0] == '\0' || s[0] == '\n')
	continue;

      i = sscanf (s, "%lx ; %[^; ]", &c, tp);
      if (i == 2)
	c2 = c;
      else
	i = sscanf (s, "%lx..%lx ; %[^; ]", &c, &c2, tp) - 1;

      if (i != 2 || c > c2 || c2 >= UNICODE_CHARS)
	die3 ("invalid input at line %ld: %s", l, s);

      typ = get_type (tp);
      for (; c <= c2; c++)
	table[c] = typ;
    }
}

static void
read_data (
  const char *data_file_type,
  const char *data_file_name
)
{
  FILE *f;

  fprintf (stderr, "Reading `%s'\n", data_file_name);
  if (!(f = fopen (data_file_name, "rt")))
    die2 ("error: cannot open `%s' for reading", data_file_name);

  if (!strcmp (data_file_type, "Scripts.txt"))
    read_scripts_txt (f);
  else
    die2 ("error: unknown data-file type %s", data_file_type);

  fclose (f);
}

static void
gen_bidi_type_tab (
  int max_depth,
  const char *data_file_type
)
{
  fprintf (stderr,
	   "Generating `" outputname "', it may take up to a few minutes\n");
  printf ("/* " outputname "\n * generated by " appname
	  "\n" " * from the file %s"". */\n\n", data_file_type);

  printf ("#define PACKTAB_UINT8 guint8\n"
	  "#define PACKTAB_UINT16 guint16\n"
	  "#define PACKTAB_UINT32 guint32\n\n");

  if (!pack_table
      (table+DIRECTLEN, UNICODE_CHARS-DIRECTLEN, 1, 0, max_depth, 32, names,
       "unsigned char", table_name, macro_name, stdout))
    die ("error: insufficient memory, decrease max_depth");

  printf ("#undef PACKTAB_UINT8\n"
	  "#undef PACKTAB_UINT16\n" "#undef PACKTAB_UINT32\n\n");

  printf ("/* End of generated " outputname " */\n");
}

int
main (
  int argc,
  const char **argv
)
{
  const char *data_file_type = "Scripts.txt";

  if (argc < 3)
    die2 ("usage:\n  " appname " max-depth /path/to/%s [junk...]",
	  data_file_type);

  {
    int max_depth = atoi (argv[1]);
    const char *data_file_name = argv[2];

    if (max_depth < 2)
      die ("invalid depth");

    init ();
    read_data (data_file_type, data_file_name);
    gen_bidi_type_tab (max_depth, data_file_type);
  }

  return 0;
}
